// Copyright 2016 Proyectos y Sistemas de Mantenimiento SL (eProsima).
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @file ReceivedChanges.h
 *
 */

#ifndef RECEIVEDCHANGES_H_
#define RECEIVEDCHANGES_H_

#include "fastrtps/rtps/common/SequenceNumber.h"

namespace eprosima {
namespace fastrtps {
namespace rtps{

class ReceivedChanges {
public:
	ReceivedChanges();
	virtual ~ReceivedChanges();

	bool add(SequenceNumber_t& s);
private:
	bool insertMissingSequenceNumbers(SequenceNumber_t& s);
	bool missingSequenceNumberReceived(SequenceNumber_t& s);

	std::vector<SequenceNumber_t> m_missingChanges;
	SequenceNumber_t m_largestSequenceNumberReceived;

};

}
} /* namespace rtps */
} /* namespace eprosima */

#endif /* RECEIVEDCHANGES_H_ */
