/*
 * Copyright 2019, Proyectos y Sistemas de Mantenimiento SL (eProsima).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef EPROSIMA_DDS_CORE_DETAIL_MODULE_DOCS_HPP_
#define EPROSIMA_DDS_CORE_DETAIL_MODULE_DOCS_HPP_

/**
 * @file
 * This file exists only to be fed to doxygen. There is quite
 * literally nothing else to see here.
 */

/**
 * @addtogroup isocpp2_dcps ISO/IEC C++ 2 API for DDS Data-centric Publish-Subscribe (DCPS)
 */
/** @{*/
/** @dir dds */
/** @}*/

/**
 * @addtogroup isocpp2_dcps_domain Domain Module - ::dds::domain
 * @ingroup isocpp2_dcps */
/** @{*/
/** @dir dds/domain */
/** @}*/


/**
 * @addtogroup isocpp2_dcps_topic Topic Module- ::dds::topic
 * @ingroup isocpp2_dcps */
/** @{*/
/** @dir dds/topic */
/** @}*/

/**
 * @addtogroup isocpp2_dcps_pub Publication Module - ::dds::pub
 * @ingroup isocpp2_dcps */
/** @{*/
/** @dir dds/pub */
/** @}*/


/**
 * @addtogroup isocpp2_dcps_sub Subscription Module - ::dds::sub
 * @ingroup isocpp2_dcps */
/** @{*/
/** @dir dds/sub */
/** @}*/

/**
 * @addtogroup isocpp2_dcps_core Core / Common Module - ::dds::core
 * @ingroup isocpp2_dcps */
/** @{*/
/** @dir dds/core */
/** @}*/


/**
 * @addtogroup examplesdcpsisocpp ISO/IEC C++ DCPS PSM Examples
 */
/** @{*/
/** @dir examples/common */
/** @}*/



/**
 * @ingroup isocpp2_dcps
 */
namespace dds {
/**
 * @ingroup isocpp2_dcps_domain
 * @ingroup isocpp2_dcps */
namespace domain {
/** @ingroup isocpp2_dcps_domain
 * @ingroup isocpp2_dcps */
namespace qos {
namespace detail {}
} //namespace qos
namespace detail {}
} //namespace domain
/**
 * @ingroup isocpp2_dcps_topic
 * @ingroup isocpp2_dcps */
namespace topic {
/** @ingroup isocpp2_dcps_topic
 * @ingroup isocpp2_dcps */
namespace qos {
namespace detail {}
} //namespace qos
namespace detail {}
} //namespace topic
/**
 * @ingroup isocpp2_dcps_pub
 * @ingroup isocpp2_dcps */
namespace pub {
/** @ingroup isocpp2_dcps_pub
 * @ingroup isocpp2_dcps */
namespace qos {
namespace detail {}
} //namespace qos
namespace detail {}
} //namespace pub
/**
 * @ingroup isocpp2_dcps_sub
 * @ingroup isocpp2_dcps */
namespace sub {
/** @ingroup isocpp2_dcps_sub
 * @ingroup isocpp2_dcps */
namespace cond {
namespace detail {}
} //namespace cond
/** @ingroup isocpp2_dcps_sub
 * @ingroup isocpp2_dcps */
namespace functors {
namespace detail {}
} //namespace functors
/** @ingroup isocpp2_dcps_sub
 * @ingroup isocpp2_dcps */
namespace status {}
/** @ingroup isocpp2_dcps_sub
 * @ingroup isocpp2_dcps */
namespace qos {
namespace detail {}
} //namespace qos
} //namespace sub
/**
 * @ingroup isocpp2_dcps_core
 * @ingroup isocpp2_dcps */
namespace core {
/** @ingroup isocpp2_dcps_core
 * @ingroup isocpp2_dcps */
namespace policy {
namespace detail {}
} //namespace policy
/** @ingroup isocpp2_dcps_core
 * @ingroup isocpp2_dcps */
namespace cond {
namespace detail {}
} //namespace cond
/** @ingroup isocpp2_dcps_core
 * @ingroup isocpp2_dcps */
namespace status {
namespace detail {}
} //namespace status
namespace detail {}
} //namespace core
} //namespace dds


#endif //EPROSIMA_DDS_CORE_DETAIL_MODULE_DOCS_HPP_
