#include <fastdds/rtps/flowcontrol/FlowControllerConsts.hpp>

namespace eprosima {
namespace fastdds {
namespace rtps {

const char* const FASTDDS_FLOW_CONTROLLER_DEFAULT = "FastDDSFlowControllerDefault";
const char* const FASTDDS_STATISTICS_FLOW_CONTROLLER_DEFAULT = "FastDDSStatisticsFlowControllerDefault";

} // namespace rtps
} // namespace fastdds
} // namespace eprosima
