// Copyright 2016 Proyectos y Sistemas de Mantenimiento SL (eProsima).
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/*! 
 * @file WideEnumTypeObject.h
 * This header file contains the declaration of the described types in the IDL file.
 *
 * This file was generated by the tool gen.
 */

#ifndef _WIDEENUM_TYPE_OBJECT_H_
#define _WIDEENUM_TYPE_OBJECT_H_


#include <fastrtps/types/TypeObject.h>
#include <map>

#if defined(_WIN32)
#if defined(EPROSIMA_USER_DLL_EXPORT)
#define eProsima_user_DllExport __declspec( dllexport )
#else
#define eProsima_user_DllExport
#endif
#else
#define eProsima_user_DllExport
#endif

#if defined(_WIN32)
#if defined(EPROSIMA_USER_DLL_EXPORT)
#if defined(WideEnum_SOURCE)
#define WideEnum_DllAPI __declspec( dllexport )
#else
#define WideEnum_DllAPI __declspec( dllimport )
#endif // WideEnum_SOURCE
#else
#define WideEnum_DllAPI
#endif
#else
#define WideEnum_DllAPI
#endif // _WIN32

using namespace eprosima::fastrtps::types;

eProsima_user_DllExport void registerWideEnumTypes();

eProsima_user_DllExport const TypeIdentifier* GetMyEnumWideIdentifier(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMyEnumWideObject(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMinimalMyEnumWideObject();
eProsima_user_DllExport const TypeObject* GetCompleteMyEnumWideObject();

eProsima_user_DllExport const TypeIdentifier* GetMyEnumWideStructIdentifier(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMyEnumWideStructObject(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMinimalMyEnumWideStructObject();
eProsima_user_DllExport const TypeObject* GetCompleteMyEnumWideStructObject();

eProsima_user_DllExport const TypeIdentifier* GetSimpleWideUnionIdentifier(bool complete = false);
eProsima_user_DllExport const TypeObject* GetSimpleWideUnionObject(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMinimalSimpleWideUnionObject();
eProsima_user_DllExport const TypeObject* GetCompleteSimpleWideUnionObject();

eProsima_user_DllExport const TypeIdentifier* GetSimpleWideUnionStructIdentifier(bool complete = false);
eProsima_user_DllExport const TypeObject* GetSimpleWideUnionStructObject(bool complete = false);
eProsima_user_DllExport const TypeObject* GetMinimalSimpleWideUnionStructObject();
eProsima_user_DllExport const TypeObject* GetCompleteSimpleWideUnionStructObject();


#endif // _WIDEENUM_TYPE_OBJECT_H_