/*
 * Copyright 2020, Proyectos y Sistemas de Mantenimiento SL (eProsima).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef OMG_DDS_QOS_DATA_READER_QOS_HPP_
#define OMG_DDS_QOS_DATA_READER_QOS_HPP_

#include <dds/sub/qos/detail/DataReaderQos.hpp>

namespace dds {
namespace sub {
namespace qos {

typedef detail::DataReaderQos DataReaderQos;

} //namespace qos
} //namespace sub
} //namespace dds

#endif //OMG_DDS_QOS_DATA_READER_QOS_HPP_
